import csv
import os
import inspect
from threading import Lock
from harnice import fileio, state

COLUMNS = [
    "net", # the physical harness (represented by a net in Kicad) that this instance is part of
    "instance_name", # the unique name of this instance
    "print_name", # the non-unique, human-readable name of this instance, used for printing on output documents
    "bom_line_number", # if this instance represents a physical procurable good, it gets assigned a line number on a bill of materials
    "mfg", # manufacturer of this instance
    "mpn",  # manufacturer part number
    "item_type",  # connector, backshell, whatever
    "parent_instance",  # general purpose reference
    "location_type",  # each instance is either better represented by one or ther other
    "segment_group",  # the group of segments that this instance is part of
    "segment_order",  # the sequential id of this item in its segment group
    "connector_group",  # a group of co-located parts (connectors, backshells, nodes)
    "channel_group", # other instances associated with this one because they are part of the same channel will share this value
    "circuit_id",  # which signal this component is electrically connected to
    "circuit_port_number",  # the sequential id of this item in its signal chain
    "node_at_end_a",  # derived from formboard definition
    "node_at_end_b",  # derived from formboard definition
    "print_name_at_end_a", # human-readable name of this instance if needed, associated with 'node_at_end_a'
    "print_name_at_end_b", # human-readable name of this instance if needed, associated with 'node_at_end_b'
    "parent_csys_instance_name",  # the other instance upon which this instance's location is based
    "parent_csys_outputcsys_name",  # the specific output coordinate system of the parent that this instance's location is based
    "translate_x",  # derived from parent_csys and parent_csys_name
    "translate_y",  # derived from parent_csys and parent_csys_name
    "rotate_csys",  # derived from parent_csys and parent_csys_name
    "absolute_rotation",  # manual add, not nominally used unless it's a flagnote, segment, or node
    "csys_children",  # imported csys children from library attributes file
    "cable_group", # other instances associated with this one because they are part of the same cable will share this value
    "cable_container", # which cable is this instance physically bundled inside of
    "cable_identifier", # cable unique identifier
    "length",  # derived from formboard definition, the length of a segment
    "diameter",  # apparent diameter of a segment <---------- change to print_diameter
    "appearance",  # see harnice.utils.appearance for details
    "note_type",  # build_note, rev_note, etc
    "note_number",  # if there is a counter involved (rev, bom, build_note, etc)
    "note_parent",  # the instance the note applies to. typically don't use this in the instances list, just note_utils
    "note_text",  # the content of the note
    "note_affected_instances",  # list of instances that are affected by the note
    "lib_repo", # publically-traceable URL of the library this instance is from
    "lib_subpath", # path to the instance within the library (directories between the product type and the part number)
    "lib_desc", # description of the instance per the library's revision history
    "lib_latest_rev", # the latest revision of the instance that exists in the remote library
    "lib_rev_used_here", # the revision of the instance that is currently used in this project
    "lib_status", # the status of the instance per the library's revision history
    "lib_releaseticket", #documentation needed
    "lib_datestarted", # the date this instance was first added to the library
    "lib_datemodified", # the date this instance was last modified in the library
    "lib_datereleased", # the date this instance was released in the library, if applicable, per the library's revision history
    "lib_drawnby", # the name of the person who drew the instance, per the library's revision history
    "lib_checkedby", # the name of the person who checked the instance, per the library's revision history
    "project_editable_lib_modified", # a flag to indicate if the imported contents do not match the library's version (it's been locally modified)
    "lib_build_notes", # recommended build notes that come with the instance from the library
    "lib_tools", # recommended tools that come with the instance from the library
    "this_instance_mating_device_refdes",  # if connector, refdes of the device it plugs into
    "this_instance_mating_device_connector",  # if connector, name of the connector it plugs into
    "this_instance_mating_device_connector_mpn",  # if connector, mpn of the connector it plugs into
    "this_net_from_device_refdes", # if this instance is a channel, circuit, conductor, etc, the refdes of the device it interfaces with, just within this net
    "this_net_from_device_channel_id", # if this instance is a channel, circuit, conductor, etc, the channel id in the device it interfaces with, just within this net
    "this_net_from_device_connector_name", # if this instance is a channel, circuit, conductor, etc, the name of the connector it interfaces with, just within this net
    "this_net_to_device_refdes", # if this instance is a channel, circuit, conductor, etc, the refdes of the device it plugs into just within this net
    "this_net_to_device_channel_id", # if this instance is a channel, circuit, conductor, etc, the channel id in the device it plugs into, just within this net
    "this_net_to_device_connector_name", # if this instance is a channel, circuit, conductor, etc, the name of the connector it plugs into, just within this net
    "this_channel_from_device_refdes",  # if this instance is a channel, circuit, conductor, etc, the refdes of the device it interfaces with, at the very end of the channel
    "this_channel_from_device_channel_id", # if this instance is a channel, circuit, conductor, etc, the channel id in the device it interfaces with, at the very end of the channel
    "this_channel_to_device_refdes",  # if this instance is a channel, circuit, conductor, etc, the refdes of the device it plugs into, at the very end of the channel
    "this_channel_to_device_channel_id", # if this instance is a channel, circuit, conductor, etc, the channel id in the device it plugs into, at the very end of the channel
    "this_channel_from_channel_type", # if this instance is a channel, circuit, conductor, etc, the type of the channel it interfaces with, at the very end of the channel
    "this_channel_to_channel_type", # if this instance is a channel, circuit, conductor, etc, the type of the channel it plugs into, at the very end of the channel
    "signal_of_channel_type", # if this instance is a channel, circuit, conductor, etc, the signal of the channel it interfaces with, at the very end of the channel
    "debug", # the call chain of the function that last modified this instance row
    "debug_cutoff", # blank cell to visually cut off the previous column
]


def new_instance(instance_name, instance_data, ignore_duplicates=False):
    """
    New Instance

instances_list.new_instance(
    instance_name,
    instance_data,
    ignore_duplicates=False
)

Add a new instance to your instances list.

    instance_name is a string and must be unique within the list.
    instance_data is a dictionary of columns (above). You may or may not include instance_name in this dict, though if you do and it doesn't match the argument, the code will fail.
    Setting ignore_duplicates to True will cause the line to pass silently if you try to add an instance with an instance_name that already exists. By default, False, if you do this, the code will raise an error if you try to add a duplicate instance_name.

    Args:
        instance_name: string, must be unique within the list
        instance_data: dictionary of columns (above)
        ignore_duplicates: boolean, default False

    Returns:
        -1 if the instance was added successfully, otherwise raises an error
    """
    if instance_name in ["", None]:
        raise ValueError(
            "Argument 'instance_name' is blank and reqired to idenitify a unique instance"
        )

    if (
        "instance_name" in instance_data
        and instance_data["instance_name"] != instance_name
    ):
        raise ValueError(
            f"Inconsistent instance_name: argument='{instance_name}' vs data['instance_name']='{instance_data['instance_name']}'"
        )

    if any(
        row.get("instance_name") == instance_name
        for row in fileio.read_tsv("instances list")
    ):
        if not ignore_duplicates:
            raise ValueError(
                f"An instance with the name '{instance_name}' already exists"
            )
        else:
            return -1

    if instance_data.get("net") is None:
        try:
            instance_data["net"] = state.net
        except AttributeError:  # no net has been set
            pass

    # Add debug call chain
    instance_data["debug"] = _get_call_chain_str()
    instance_data["debug_cutoff"] = " "

    # add argumet to data added
    instance_data["instance_name"] = instance_name

    with open(fileio.path("instances list"), "a", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writerow({key: instance_data.get(key, "") for key in COLUMNS})


_instances_lock = Lock()


def modify(instance_name, instance_data):
    with _instances_lock:
        path = fileio.path("instances list")

        # --- Read once ---
        with open(path, newline="", encoding="utf-8") as f:
            reader = csv.DictReader(f, delimiter="\t")
            rows = list(reader)
            fieldnames = reader.fieldnames or []

        # --- Add debug info before updating ---
        instance_data["debug"] = _get_call_chain_str()
        instance_data["debug_cutoff"] = " "

        # Ensure any new keys are part of the header
        for key in instance_data:
            if key not in fieldnames:
                fieldnames.append(key)

        # --- Modify in-place ---
        found = False
        for row in rows:
            if row.get("instance_name") == instance_name:
                row.update(instance_data)
                found = True
                break

        if not found:
            raise ValueError(f"Instance '{instance_name}' not found")

        # --- Write atomically ---
        tmp = path + ".tmp"
        with open(tmp, "w", newline="", encoding="utf-8") as f:
            writer = csv.DictWriter(f, fieldnames=fieldnames, delimiter="\t")
            writer.writeheader()
            writer.writerows(rows)
            f.flush()
            os.fsync(f.fileno())

        os.replace(tmp, path)


def remove_instance(instance_to_delete):
    instances_list = fileio.read_tsv("instances list")
    new_list = []
    for instance in instances_list:
        if instance.get("instance_name") == instance_to_delete.get("instance_name"):
            continue
        new_list.append(instance)

    with open(fileio.path("instances list"), "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows(new_list)


def new():
    with open(fileio.path("instances list"), "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows([])


def assign_bom_line_numbers():
    bom = []
    for instance in fileio.read_tsv("instances list"):
        if instance.get("bom_line_number") == "True":
            if instance.get("mpn") == "":
                raise ValueError(
                    f"You've chosen to add {instance.get('instance_name')} to the bom, but haven't specified an MPN"
                )
            if instance.get("mpn") not in bom:
                bom.append(instance.get("mpn"))

    bom_line_number = 1
    for bom_item in bom:
        for instance in fileio.read_tsv("instances list"):
            if instance.get("mpn") == bom_item:
                modify(
                    instance.get("instance_name"), {"bom_line_number": bom_line_number}
                )
        bom_line_number += 1


def attribute_of(target_instance, attribute):
    for instance in fileio.read_tsv("instances list"):
        if instance.get("instance_name") == target_instance:
            return instance.get(attribute)


def instance_in_connector_group_with_item_type(connector_group, item_type):
    if connector_group in ["", None]:
        raise ValueError("Connector group is blank")
    if item_type in ["", None]:
        raise ValueError("Suffix is blank")
    match = 0
    output = None
    for instance in fileio.read_tsv("instances list"):
        if instance.get("connector_group") == connector_group:
            if instance.get("item_type") == item_type:
                match = match + 1
                output = instance
    if match == 0:
        return 0
    if match > 1:
        raise ValueError(
            f"Multiple instances found in connector_group '{connector_group}' with item type '{item_type}'."
        )
    return output


def _get_call_chain_str():
    """
    Returns the call chain as a readable string:
    filename:line in function -> filename:line in function ...
    """
    stack = inspect.stack()
    chain_parts = []
    for frame_info in reversed(stack[1:]):  # skip this function itself
        filename = os.path.basename(frame_info.filename)
        lineno = frame_info.lineno
        function = frame_info.function
        chain_parts.append(f"{filename}:{lineno} in {function}()")
    return " -> ".join(chain_parts)


def list_of_uniques(attribute):
    output = []
    for instance in fileio.read_tsv("instances list"):
        if instance.get(attribute) not in output:
            if instance.get(attribute) not in [None, ""]:
                output.append(instance.get(attribute))
    return output



import os
import csv
from harnice import fileio

COLUMNS = [
    "A-side_device_refdes", #documentation needed
    "A-side_device_channel_id", #documentation needed
    "A-side_device_channel_type", #documentation needed
    "B-side_device_refdes", #documentation needed
    "B-side_device_channel_id", #documentation needed
    "B-side_device_channel_type", #documentation needed
    "disconnect_refdes", #documentation needed
    "disconnect_channel_id", #documentation needed
    "A-port_channel_type", #documentation needed
    "B-port_channel_type", #documentation needed
    "manual_map_channel_python_equiv", #documentation needed
]


def new():
    disconnect_map_rows = []

    for channel in fileio.read_tsv("channel map"):
        raw = (channel.get("disconnect_refdes_requirement") or "").strip()
        if not raw:
            continue

        # split on semicolon -> one row per disconnect_refdes requirement
        disconnects = [item.strip() for item in raw.split(";") if item.strip()]

        for requirement in disconnects:
            # requirement looks like "X1(A,B)" or "X2(B,A)"
            refdes, ports = requirement.split("(")
            ports = ports.rstrip(")")
            first_port, second_port = [p.strip() for p in ports.split(",")]

            # orientation: (A,B) means from_device is A-side, (B,A) means from_device is B-side
            if (first_port, second_port) == ("A", "B"):
                a_refdes = channel.get("from_device_refdes", "")
                a_chan_id = channel.get("from_device_channel_id", "")
                a_chan_type_id = channel.get("from_channel_type", "")
                b_refdes = channel.get("to_device_refdes", "")
                b_chan_id = channel.get("to_device_channel_id", "")
                b_chan_type_id = channel.get("to_channel_type", "")
            elif (first_port, second_port) == ("B", "A"):
                b_refdes = channel.get("from_device_refdes", "")
                b_chan_id = channel.get("from_device_channel_id", "")
                b_chan_type_id = channel.get("from_channel_type", "")
                a_refdes = channel.get("to_device_refdes", "")
                a_chan_id = channel.get("to_device_channel_id", "")
                a_chan_type_id = channel.get("to_channel_type", "")
            else:
                raise ValueError(f"Unexpected port order: {requirement}")

            disconnect_map_rows.append(
                {
                    "A-side_device_refdes": a_refdes,
                    "A-side_device_channel_id": a_chan_id,
                    "A-side_device_channel_type": a_chan_type_id,
                    "B-side_device_refdes": b_refdes,
                    "B-side_device_channel_id": b_chan_id,
                    "B-side_device_channel_type": b_chan_type_id,
                    "disconnect_refdes": refdes.strip(),
                }
            )

    for item in fileio.read_tsv("bom"):
        if item.get("disconnect"):
            disconnect_signals_list_path = os.path.join(
                fileio.dirpath("instance_data"),
                "disconnect",
                item.get("device_refdes"),
                f"{item.get('device_refdes')}-signals_list.tsv",
            )

            available_disconnect_channels = set()
            for signal in fileio.read_tsv(disconnect_signals_list_path):
                if signal.get("channel_id") in available_disconnect_channels:
                    continue
                available_disconnect_channels.add(signal.get("channel_id"))

                disconnect_map_rows.append(
                    {
                        "disconnect_refdes": item.get("device_refdes"),
                        "disconnect_channel_id": signal.get("channel_id"),
                        "A-port_channel_type": signal.get("A_channel_type"),
                        "B-port_channel_type": signal.get("B_channel_type"),
                    }
                )

    with open(fileio.path("disconnect map"), "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows(disconnect_map_rows)

    # initialize mapped disconnect channels set (empty TSV)
    with open(
        fileio.path("mapped disconnects set"), "w", newline="", encoding="utf-8"
    ) as f:
        pass
    with open(
        fileio.path("mapped A-side channels through disconnects set"),
        "w",
        newline="",
        encoding="utf-8",
    ) as f:
        pass


def assign(a_side_key, disconnect_key):
    # a_side is the (device refdes, channel_id) that is on the A-side of the disconnect
    channels = fileio.read_tsv("disconnect map")
    if channel_is_already_assigned_through_disconnect(a_side_key, disconnect_key[0]):
        raise ValueError(f"disconnect_key {disconnect_key} already assigned")

    if disconnect_is_already_assigned(disconnect_key):
        raise ValueError(f"disconnect {disconnect_key} already assigned")

    # Find the disconnect row we want to merge
    disconnect_info = None
    for row in channels:
        if (
            row.get("disconnect_refdes") == disconnect_key[0]
            and row.get("disconnect_channel_id") == disconnect_key[1]
            and row.get("A-side_device_refdes") in [None, ""]
        ):
            disconnect_info = row
            break

    updated_channels = []
    for row in channels:
        if (
            row.get("A-side_device_refdes") == a_side_key[0]
            and row.get("A-side_device_channel_id") == a_side_key[1]
            and row.get("disconnect_refdes") == disconnect_key[0]
        ):
            row["disconnect_channel_id"] = disconnect_key[1]
            row["A-port_channel_type"] = disconnect_info.get("A-port_channel_type", "")
            row["B-port_channel_type"] = disconnect_info.get("B-port_channel_type", "")
            row["manual_map_channel_python_equiv"] = (
                f"disconnect_map.assign({a_side_key}, {disconnect_key})"
            )

        elif (
            row.get("disconnect_refdes") == disconnect_key[0]
            and row.get("disconnect_channel_id") == disconnect_key[1]
            and row.get("A-side_device_refdes") in [None, ""]
        ):
            continue

        updated_channels.append(row)

    already_assigned_channels_through_disconnects_set_append(
        a_side_key, disconnect_key[0]
    )
    already_assigned_disconnects_set_append(disconnect_key)

    with open(fileio.path("disconnect map"), "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows(updated_channels)


def already_assigned_channels_through_disconnects_set_append(key, disconnect_refdes):
    item = f"{key}:{disconnect_refdes}"
    items = set(already_assigned_channels_through_disconnects_set())
    if item in items:
        raise ValueError(
            f"channel {key} through disconnect {disconnect_refdes} already assigned"
        )
    with open(
        fileio.path("mapped A-side channels through disconnects set"),
        "a",
        newline="",
        encoding="utf-8",
    ) as f:
        writer = csv.writer(f, delimiter="\t")
        writer.writerow([item])


def already_assigned_disconnects_set_append(key):
    items = set(already_assigned_disconnects_set())
    if str(key) in items:
        raise ValueError(f"disconnect {key} already assigned to a channel")
    items.add(str(key))
    with open(
        fileio.path("mapped disconnects set"), "w", newline="", encoding="utf-8"
    ) as f:
        writer = csv.writer(f, delimiter="\t")
        for item in sorted(items):
            writer.writerow([item])


def already_assigned_channels_through_disconnects_set():
    items = []
    with open(
        fileio.path("mapped A-side channels through disconnects set"),
        newline="",
        encoding="utf-8",
    ) as f:
        reader = csv.reader(f, delimiter="\t")
        for row in reader:
            if row and row[0].strip():  # skip blank lines
                items.append(row[0].strip())
    return items


def already_assigned_disconnects_set():
    items = []
    with open(fileio.path("mapped disconnects set"), newline="", encoding="utf-8") as f:
        reader = csv.reader(f, delimiter="\t")
        for row in reader:
            if row and row[0].strip():
                items.append(row[0].strip())
    return items


def channel_is_already_assigned_through_disconnect(key, disconnect_refdes):
    if (
        f"{str(key)}:{disconnect_refdes}"
        in already_assigned_channels_through_disconnects_set()
    ):
        return True
    else:
        return False


def disconnect_is_already_assigned(key):
    if str(key) in already_assigned_disconnects_set():
        return True
    else:
        return False

def ensure_requirements_met():
    for row in fileio.read_tsv("disconnect map"):
        if row.get("A-side_device_refdes") not in [None, ""] and row.get("disconnect_channel_id") in [None, ""]:
            raise ValueError(f"Channel '{row.get('A-side_device_refdes')}.{row.get('A-side_device_channel_id')}' to '{row.get('B-side_device_refdes')}.{row.get('B-side_device_channel_id')}' could not find a compatible disconnect channel through '{row.get('disconnect_refdes')}'")


import os
import re
import shutil
import filecmp
import json
from harnice import fileio
from harnice.lists import instances_list, library_history, rev_history
from harnice.cli import print_import_status

"""
where a part lands in a project after it's been imported:

instance_data
    item_type
        destination_directory
            lib_used
                lib_used_rev

"""


def pull(input_dict, update_instances_list=True, destination_directory=None):
    """
    Imports a part from the library into the project.

    Copies a part (device, connector, cable, etc.) from the library repository into
    the project's `instance_data` directory. Handles revision selection, file copying,
    and updating the instances list and library history. The function:

    1. Validates required fields (`lib_repo`, `mpn`, `item_type`)
    2. Determines which revision to use (specified or latest available)
    3. Copies the library revision to `library_used_do_not_edit`
    4. Copies editable files to the instance directory (only if not already present)
    5. Updates the instances list with library metadata
    6. Records the import in library history

    **Args:**
    - `input_dict` (dict): Dictionary containing part information with required keys:
        - `instance_name` (str): Name for this instance in the project
        - `lib_repo` (str): Library repository URL or `"local"` for local library
        - `mpn` (str): Manufacturer part number
        - `item_type` (str): Type of item (device, connector, cable, etc.)
        - `lib_subpath` (str, optional): Subpath within the library
        - `lib_rev_used_here` (str, optional): Specific revision to use (e.g., `"1"` or `"rev1"`)
    - `update_instances_list` (bool, optional): If `True`, updates the instances list with
        library metadata. Defaults to `True`.
    - `destination_directory` (str, optional): Custom destination directory. If `None`,
        defaults to `instance_data/{item_type}/{instance_name}`.

    **Returns:**
    - `str`: Path to the destination directory where the part was imported.

    **Raises:**
    - `ValueError`: If required fields (`lib_repo`, `mpn`, `item_type`) are blank.
    - `FileNotFoundError`: If no revision folders are found for the part number in the library.
    """
    # throw errors if required fields are blank
    if input_dict.get("lib_repo") in [None, ""]:
        raise ValueError(
            f"when importing {input_dict.get('instance_name')} 'lib_repo' is required but blank"
        )
    if input_dict.get("mpn") in [None, ""]:
        raise ValueError(
            f"when importing {input_dict.get('instance_name')} 'mpn' is required but blank"
        )
    if input_dict.get("item_type") in [None, ""]:
        raise ValueError(
            f"when importing {input_dict.get('instance_name')} 'item_type' is required but blank"
        )

    # determine destination directory
    if destination_directory is None:
        destination_directory = os.path.join(
            fileio.dirpath(None),
            "instance_data",
            input_dict.get("item_type"),
            input_dict.get("instance_name"),
        )
    os.makedirs(destination_directory, exist_ok=True)

    lib_repo = None
    if input_dict.get("lib_repo") == "local":
        lib_repo = os.path.join(fileio.part_directory(), "library")
    else:
        lib_repo = get_local_path(input_dict.get("lib_repo"))

    # determine source library path
    source_lib_path = os.path.join(
        lib_repo,
        input_dict.get("item_type"),
        input_dict.get("lib_subpath", ""),
        input_dict.get("mpn"),
    )

    # === Find highest rev in library
    source_revision_folders = [
        name
        for name in os.listdir(source_lib_path)
        if os.path.isdir(os.path.join(source_lib_path, name))
        and re.fullmatch(
            rf"{re.escape(input_dict.get('mpn').lower())}-rev(\d+)", name.lower()
        )
    ]
    if not source_revision_folders:
        raise FileNotFoundError(
            f"No revision folders found for {input_dict.get('mpn')} in {source_lib_path}"
        )
    highest_source_rev = str(
        max(
            int(re.search(r"rev(\d+)", name).group(1))
            for name in source_revision_folders
        )
    )
    # === Decide which rev to use
    if input_dict.get("lib_rev_used_here"):
        rev_to_use = int(
            input_dict.get("lib_rev_used_here").strip().lower().replace("rev", "")
        )
        if int(highest_source_rev) > int(rev_to_use):
            import_state = f"newer rev exists   (rev{rev_to_use} used, rev{highest_source_rev} available)"
        else:
            import_state = f"library up to date (rev{rev_to_use})"
    else:
        rev_to_use = highest_source_rev
        import_state = f"imported latest (rev{rev_to_use})"

    # === Import library contents freshly every time
    lib_used_path = os.path.join(destination_directory, "library_used_do_not_edit")
    os.makedirs(lib_used_path, exist_ok=True)

    lib_used_rev_path = os.path.join(
        lib_used_path, f"{input_dict.get('mpn')}-rev{rev_to_use}"
    )
    if os.path.exists(lib_used_rev_path):
        shutil.rmtree(lib_used_rev_path)

    source_lib_rev_path = os.path.join(
        source_lib_path, f"{input_dict.get('mpn')}-rev{rev_to_use}"
    )

    shutil.copytree(source_lib_rev_path, lib_used_rev_path)

    # === Copy editable files into the editable directory only if not already present
    rename_suffixes = [
        "-drawing.svg",
        "-params.json",
        "-attributes.json",
        "-signals_list.tsv",
        "-feature_tree.py",
        "-conductor_list.tsv",
    ]
    Modified = False
    for filename in os.listdir(lib_used_rev_path):
        lib_used_do_not_edit_file = os.path.join(lib_used_rev_path, filename)

        new_name = filename
        for suffix in rename_suffixes:
            if filename.endswith(suffix):
                new_name = f"{input_dict.get('instance_name')}{suffix}"
                break

        editable_file_path = os.path.join(destination_directory, new_name)
        if not os.path.exists(editable_file_path):
            shutil.copy2(lib_used_do_not_edit_file, editable_file_path)

            # special rules for copying svg
            if new_name.endswith(".svg"):
                with open(editable_file_path, "r", encoding="utf-8") as f:
                    content = f.read()
                content = content.replace(
                    f"{input_dict.get('mpn')}-drawing-contents-start",
                    f"{input_dict.get('instance_name')}-contents-start",
                ).replace(
                    f"{input_dict.get('mpn')}-drawing-contents-end",
                    f"{input_dict.get('instance_name')}-contents-end",
                )
                with open(editable_file_path, "w", encoding="utf-8") as f:
                    f.write(content)

        else:
            # Compare the existing editable file and the library version
            if not filecmp.cmp(
                lib_used_do_not_edit_file, editable_file_path, shallow=False
            ):
                Modified = Modified or True

    if Modified:
        import_state = f"modified in this project (rev{rev_to_use})"
    else:
        import_state = f"up to date (rev{rev_to_use})"

    # === Load revision row from revision history TSV in source library ===
    revhistory_path = os.path.join(
        source_lib_path, f"{input_dict.get('mpn')}-revision_history.tsv"
    )
    revhistory_row = rev_history.info(rev=rev_to_use, path=revhistory_path)

    try:
        with open(
            os.path.join(
                destination_directory,
                f"{input_dict.get('instance_name')}-attributes.json",
            ),
            "r",
            encoding="utf-8",
        ) as f:
            attributes_data = json.load(f)

        csys_children = attributes_data.get("csys_children") or {}
        tools = attributes_data.get("tools") or []
        build_notes = attributes_data.get("build_notes") or []

    except Exception:
        csys_children = {}
        tools = []
        build_notes = []

    update_contents = {
        "mpn": input_dict.get("mpn"),
        "item_type": input_dict.get("item_type"),
        "csys_children": csys_children,
        "lib_repo": lib_repo,
        "lib_subpath": input_dict.get("lib_subpath"),
        "lib_desc": revhistory_row.get("desc"),
        "lib_latest_rev": highest_source_rev,
        "lib_rev_used_here": rev_to_use,
        "lib_status": revhistory_row.get("status"),
        "lib_releaseticket": revhistory_row.get("releaseticket"),
        "lib_datestarted": revhistory_row.get("datestarted"),
        "lib_datemodified": revhistory_row.get("datemodified"),
        "lib_datereleased": revhistory_row.get("datereleased"),
        "lib_drawnby": revhistory_row.get("drawnby"),
        "lib_checkedby": revhistory_row.get("checkedby"),
        "lib_tools": tools,
        "lib_build_notes": build_notes,
        "project_editable_lib_modified": Modified,
    }

    if update_instances_list:
        try:
            instances_list.modify(input_dict.get("instance_name"), update_contents)
        except ValueError:
            instances_list.new_instance(
                input_dict.get("instance_name"), update_contents
            )

    library_history.append(input_dict.get("instance_name"), update_contents)

    print_import_status(
        input_dict.get("instance_name"),
        input_dict.get("item_type"),
        update_contents.get("lib_status"),
        import_state,
        os.path.basename(
            os.path.dirname(os.path.dirname(os.path.dirname(destination_directory)))
        ),
    )
    return destination_directory


def get_local_path(lib_repo):
    """
    Looks up the local filesystem path for a library repository URL.

    Reads the `library_locations.csv` file to find the mapping between a library
    repository URL and its local filesystem path. If the CSV file doesn't exist,
    it creates one with a default entry for the `harnice-library-public` repository.

    The lookup is case-insensitive. The local path is expanded (e.g., `~` is expanded
    to the user's home directory).

    **Args:**
    - `lib_repo` (str): Library repository URL to look up (e.g.,
        `"https://github.com/harnice/harnice"`).

    **Returns:**
    - `str`: Local filesystem path to the library repository.

    **Raises:**
    - `ValueError`: If the library repository URL is not found in the CSV file,
        or if no local path is specified for the repository.
    """
    csv_path = fileio.path("library locations")  # path to library_locations.csv

    # ----------------------------------------------------
    # If file does not exist, auto-generate with default
    # ----------------------------------------------------
    if not os.path.exists(csv_path):
        # Determine base directory of the Harnice repo
        # (__file__) → .../harnice/utils/library_utils.py
        # dirname 3 times → repo root
        repo_root = os.path.dirname(
            os.path.dirname(
                os.path.dirname(
                    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
                )
            )
        )

        default_local_path = os.path.join(repo_root, "harnice-library-public")

        # Ensure the directory exists for the CSV
        os.makedirs(os.path.dirname(csv_path), exist_ok=True)

        with open(csv_path, "w", encoding="utf-8") as f:
            f.write(f"https://github.com/harnice/harnice,{default_local_path}\n")

        print(f"[harnice] Created '{csv_path}'")
        print(f"[harnice] Default library-public location: {default_local_path}")

    # ----------------------------------------------------
    # Normal lookup (with BOM-safe decoding)
    # ----------------------------------------------------
    with open(csv_path, newline="", encoding="utf-8-sig") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            parts = line.split(",")
            if len(parts) < 2:
                continue

            url = parts[0].strip()
            local = parts[1].strip()

            # Case-insensitive match
            if url.lower() == lib_repo.lower().strip():
                if not local:
                    raise ValueError(f"No local path found for '{lib_repo}'")
                return os.path.expanduser(local)

    raise ValueError(f"'{lib_repo}' not found in library locations")
